from textwrap import dedent
from unittest import TestCase
from unittest.mock import patch

from botocore.exceptions import NoRegionError
from botocore.stub import Stubber

from tests.recruitment.agency import fake_credentials
from tests.recruitment.agency import client
from tests.recruitment.agency import raise_this
from recruitment.agency import Broker
from recruitment.agency import Config
from recruitment.agency import Communicator


class CommunicatorTest(TestCase):

    @patch('boto3.client')
    def test_cannot_instantiate_with_invalid_Config(self, mock_boto_client):
        mock_boto_client.side_effect = raise_this(exception=ValueError)
        with self.assertRaises(Communicator.FailedToInstantiate):
            Communicator(config=Config(Broker.sns))

        mock_boto_client.side_effect = raise_this(exception=NoRegionError)
        with self.assertRaises(Communicator.FailedToInstantiate):
            Communicator(config=Config(Broker.sns))

    @patch('boto3.client')
    def test_communicator_instantiation_failure_redacts_secrets(self, mock_boto_client):
        expected_serialization = dedent(
            """\
            service_name=sns
            region_name=
            aws_access_key_id=**********
            aws_secret_access_key=**********
            endpoint_url="""
        )

        mock_boto_client.side_effect = raise_this(exception=ValueError)
        with self.assertRaises(Communicator.FailedToInstantiate) as error_ctx:
            Communicator(config=Config(Broker.sns))

        self.assertEqual(str(error_ctx.exception), expected_serialization)

    @patch('boto3.client')
    def test_can_create_topic_and_publish_message(self, mock_boto_client):
        region = 'some-region-1'
        topic_name = 'some-topic'
        broker = Broker.sns

        expected_topic_response = {
            'TopicArn': f'arn:aws:sns:{region}:12345:{topic_name}'
        }
        expected_publish_response = {
            'MessageId': '00000000-0000-0000-0000-000000000000'
        }

        sns = client(broker.name, region)
        mock_boto_client.return_value = sns
        with Stubber(sns) as stubber:
            stubber.add_response(
                broker.interface['declare_receiver'], expected_topic_response
            )
            stubber.add_response(broker.interface['send'], expected_publish_response)

            commlink = Communicator(Config(broker, **fake_credentials))
            topic_creation = commlink.declare_receiver(
                Name='string',
                Attributes={'string': 'string'},
                Tags=[
                    {'Key': 'string', 'Value': 'string'},
                ],
            )
            message_receipt = commlink.send(Message='some message!')

        self.assertEqual(topic_creation, expected_topic_response)
        self.assertEqual(message_receipt, expected_publish_response)



from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from recruitment.agency import Consumer
from tests.recruitment.agency import uncloseable


class ConsumerTest(TestCase):

    consumer = Consumer()
    head = 'head'
    tail = 'tail'
    separator = '\n'
    head_and_tail = f'{head}{separator}{tail}'

    @patch('builtins.open')
    def test_can_take_oldest_deadletter_first(self, mock_file):
        Consumer.deadletter_file = __file__
        with uncloseable(StringIO(self.head_and_tail)) as buffer:
            mock_file.return_value = buffer
            result = self.consumer.take_deadletter()
            self.assertTrue(result.successful)
            self.assertEqual(result.value, f'{self.head}{self.separator}')

    def test_can_consume(self):
        with self.assertRaises(NotImplementedError):
            self.consumer.consume()



from unittest import TestCase
from unittest.mock import patch

from recruitment.agency import Config
from recruitment.agency import Publisher
from recruitment.agency.protocols import HasContingency

from tests.recruitment.agency import acceptable_broker_names


class HasContingencyTest(TestCase):

    @patch('boto3.client')
    def test_can_follow_protocol(self, mock_boto_client):
        broker_name = acceptable_broker_names[0]
        self.assertIsInstance(Publisher(Config(broker_name)), HasContingency)



from contextlib import contextmanager
from io import IOBase as Buffer

from botocore.session import get_session


acceptable_broker_names = (
    'kinesis',
    's3',
    'sns',
    'sqs',
)

fake_credentials = {
    'region_name': 'somewhere-in-the-world',
    'aws_access_key_id': 's3curityBadge!',
    'aws_secret_access_key': 'p@ssw0rd!',
    'endpoint_url': 'some-computer.com',
}


def client(client_type: str, region_name: str):
    return get_session().create_client(client_type, region_name=region_name)


def raise_this(**kwargs):
    exception = kwargs['exception']
    def will_raise(**kwargs):
        raise exception
    return will_raise


@contextmanager
def uncloseable(buffer: Buffer):
    """Context manager which makes the buffer's close operation a no-op"""
    close = buffer.close
    buffer.close = lambda: None
    yield buffer
    buffer.close = close
    buffer.seek(0)  # fake close



import boto3

from dataclasses import asdict
from dataclasses import dataclass
from enum import Enum
from enum import auto
from functools import reduce
from functools import partial
from os import environ as envvars
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

from botocore.exceptions import NoRegionError
from actionpack import Action
from actionpack.actions import Call
from actionpack.actions import Remove
from actionpack.actions import RetryPolicy
from actionpack.actions import Write
from actionpack.utils import Closure


local_storage_dir = Path.home() / '.recruitment/agency/'
deadletters = local_storage_dir / 'deadletters'

class Broker(Enum):
    """A repository for declaring services and their interfaces"""

    def _generate_next_value_(name, start, count, last_values):
        return name

    s3 = auto()
    sns = auto()
    sqs = auto()
    kinesis = auto()

    @property
    def interface(self) -> Dict[str, Optional[str]]:
        send = 'send'
        declare_receiver = 'declare_receiver'
        methods_for = {
            Broker.s3: {send: 'upload_file', declare_receiver: 'create_bucket'},
            Broker.sns: {send: 'publish', declare_receiver: 'create_topic'},
            Broker.sqs: {send: 'send_message', declare_receiver: 'create_queue'},
            Broker.kinesis: {send: 'put_record', declare_receiver: 'create_stream'},
        }
        return methods_for[self]


@dataclass
class Config:
    """An object for conveying configuration info"""

    service_name: Union[str, Broker]
    region_name: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    endpoint_url: Optional[str] = None

    @staticmethod
    def fromenv():
        return Config(
            service_name=envvars.get('AWS_SERVICE_NAME'),
            region_name=envvars.get('AWS_REGION_NAME'),
            aws_access_key_id=envvars.get('AWS_ACCESS_KEY_ID'),
            aws_secret_access_key=envvars.get('AWS_SECRET_ACCESS_KEY'),
            endpoint_url=envvars.get('AWS_ENDPOINT_URL')
        )

    def asfile(self, profile: str = 'default'):
        return f'[{profile}]\n{str(self)}'

    def __post_init__(self):
        if self.service_name is None:
            raise Config.AttributeDeclaredIncorrectly('Missing service_name.')

        if isinstance(self.service_name, Broker):
            self.service_name = self.service_name.value
        elif isinstance(self.service_name, str):
            self.service_name = Broker(self.service_name).value
        else:
            service_name_type = type(self.service_name).__name__
            raise Config.AttributeDeclaredIncorrectly(
                f'Service name must be a <Broker> or <str>. Received <{service_name_type}>.'
            )

    def __str__(self):
        """Produces a string compatible with the AWS CLI tool
        (see documentation for details:
         https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html)
        """
        return reduce(
            lambda a, b: f'{a}' + f'{b[0]}={b[1] if b[1] else ""}\n',
            asdict(self).items(),
            '',
        ).strip()

    class AttributeDeclaredIncorrectly(Exception):
        pass


class Communicator:
    """An object that hosts the Broker.interface"""

    def __init__(self, config: Config):
        broker = Broker(config.service_name)  # maybe redundant
        _client = partial(boto3.client, service_name=broker.name)
        for alias, method in broker.interface.items():
            try:
                client = _client(
                    region_name=config.region_name, endpoint_url=config.endpoint_url
                )
            except (ValueError, NoRegionError) as e:
                raise Communicator.FailedToInstantiate(given=config) from e
            setattr(self, alias, getattr(client, method))

    class FailedToInstantiate(Exception):
        def __init__(self, given: Config):
            redaction = '*' * 10
            redacted_config = Config(
                service_name=given.service_name,
                region_name=given.region_name,
                aws_access_key_id=redaction,
                aws_secret_access_key=redaction,
                endpoint_url=given.endpoint_url
            )
            super().__init__(str(redacted_config))


class Publisher:
    """A namespace for publishing messages"""

    def __init__(
        self,
        config: Config,
        retry_policy_provider: Optional[Callable[[Action], RetryPolicy]] = None,
        record_failure_provider: Optional[Callable[[], Write]] = None,
    ):
        self.communicator = Communicator(config)
        self.retry_policy_provider = retry_policy_provider
        self.record_failure_provider = record_failure_provider

    def publish(self, *args, **kwargs):
        send_communique = Call(Closure(self.communicator.send, *args, **kwargs))
        if self.retry_policy_provider:
            retry_policy = self.retry_policy_provider(send_communique)
            result = retry_policy.perform()

            if not result.successful and self.record_failure_provider:
                record_failure = self.record_failure_provider()
                record_failure.perform()

            return result, retry_policy.attempts
        else:
            return send_communique.perform()


class Consumer:
    """A namespace for consuming messages

    TODO (withtwoemms) -- add error logfile header for simpler parsing
    """

    deadletter_file = deadletters / 'consumer' / 'letters'

    def consume(self):
        """Consume from message bus"""
        raise NotImplementedError()

    def take_deadletter(self):
        _consume = Remove(filename=self.deadletter_file)
        return _consume.perform()


class Agent(Consumer, Publisher):
    """A namespace for consuming and/or publishing messages"""

    def __init__(
        self,
        config: Config,
        retry_policy_provider: Optional[Callable[[Action], RetryPolicy]] = None,
        record_failure_provider: Optional[Callable[[], Write]] = None,
    ):
        Publisher.__init__(self, config, retry_policy_provider, record_failure_provider)



from datetime import datetime
from io import StringIO
from unittest import TestCase
from unittest.mock import patch

from actionpack.actions import Call
from actionpack.actions import RetryPolicy
from actionpack.actions import Write
from actionpack.utils import Closure
from botocore.exceptions import ClientError
from botocore.stub import Stubber

from tests.recruitment.agency import client
from tests.recruitment.agency import fake_credentials
from tests.recruitment.agency import uncloseable
from recruitment.agency import Agent
from recruitment.agency import Broker
from recruitment.agency import Communicator
from recruitment.agency import Config


class AgentTest(TestCase):

    broker = Broker.sns
    region = 'some-region-1'
    sns = client(broker.name, region)
    expected_publish_response = {'MessageId': '00000000-0000-0000-0000-000000000000'}

    write_to_deadletter_file = Write(
        prefix=f'-> [{datetime.utcnow()}] -- ',
        filename=Agent.deadletter_file,
        to_write='failed',
        append=True,
        mkdir=True,
    )

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_can_retry_message_send(self, mock_write, mock_boto_client):
        mock_boto_client.return_value = self.sns
        with Stubber(self.sns) as stubber:
            stubber.add_client_error(self.broker.interface['send'], '500')  # attempt
            stubber.add_client_error(self.broker.interface['send'], '500')  # retry 1
            stubber.add_client_error(self.broker.interface['send'], '500')  # retry 2
            smith = Agent(
                config=Config(self.broker, **fake_credentials),
                retry_policy_provider=lambda action: RetryPolicy(
                    action, max_retries=2, should_record=True
                ),
            )
            result, attempts = smith.publish(Message='Mr. Anderson...')

        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, RetryPolicy.Expired)
        self.assertEqual(len(attempts), 3)
        for attempt in attempts:
            self.assertIsInstance(attempt.value, ClientError)

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_publish_can_eventually_succeed(self, mock_write, mock_boto_client):
        mock_boto_client.return_value = self.sns
        with Stubber(self.sns) as stubber:
            stubber.add_client_error(self.broker.interface['send'], '500')  # attempt
            stubber.add_client_error(self.broker.interface['send'], '500')  # retry 1
            stubber.add_response(
                self.broker.interface['send'], self.expected_publish_response
            )
            smith = Agent(
                config=Config(self.broker, **fake_credentials),
                retry_policy_provider=lambda action: retry_policy_provider(action),
            )
            result, attempts = smith.publish(Message='Mr. Anderson...')

        self.assertTrue(result.successful)
        self.assertEqual(result.value, self.expected_publish_response)
        self.assertEqual(len(attempts), 2)
        for attempt in attempts:
            self.assertIsInstance(attempt.value, ClientError)

    @patch('boto3.client')
    @patch('actionpack.actions.Write.perform')
    def test_can_define_failure_callback(self, mock_write, mock_boto_client):
        vessel = []
        contents = 'water'

        def fill(vessel: list, contents):
            vessel.append(contents)
            return vessel

        callback = Call(Closure(fill, vessel, contents))

        self.assertFalse(vessel)  # confirms the vessel has yet to be filled
        mock_boto_client.return_value = self.sns
        with Stubber(self.sns) as stubber:
            stubber.add_client_error(self.broker.interface['send'], '500')
            stubber.add_client_error(self.broker.interface['send'], '500')
            stubber.add_client_error(self.broker.interface['send'], '500')
            smith = Agent(
                config=Config(self.broker, **fake_credentials),
                retry_policy_provider=lambda action: retry_policy_provider(
                    action,
                    reaction=callback,  # called if the RetryPolicy expires
                ),
            )
            result, attempts = smith.publish(Message='Mr. Anderson...')

        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, RetryPolicy.Expired)
        self.assertEqual(len(attempts), 3)
        for attempt in attempts:
            self.assertIsInstance(attempt.value, ClientError)
        self.assertEqual(vessel, [contents])  # evidence the fill callback was called

    @patch('boto3.client')
    @patch('pathlib.Path.open')
    def test_can_write_deadletter(self, mock_file, mock_boto_client):
        mock_boto_client.return_value = self.sns
        with uncloseable(StringIO()) as buffer:
            mock_file.return_value = buffer

            with Stubber(self.sns) as stubber:
                stubber.add_client_error(self.broker.interface['send'], '500')
                stubber.add_client_error(self.broker.interface['send'], '500')
                stubber.add_client_error(self.broker.interface['send'], '500')
                smith = Agent(
                    config=Config(self.broker, **fake_credentials),
                    retry_policy_provider=lambda action: retry_policy_provider(action),
                    record_failure_provider=lambda: self.write_to_deadletter_file,
                )
                result, attempts = smith.publish(Message='Mr. Anderson...')

        buffer_contents: str = buffer.read()
        self.assertTrue(buffer_contents.startswith('->'))
        self.assertTrue(buffer_contents.endswith('failed\n'))

        self.assertFalse(result.successful)
        self.assertIsInstance(result.value, RetryPolicy.Expired)
        self.assertEqual(len(attempts), 3)
        for attempt in attempts:
            self.assertIsInstance(attempt.value, ClientError)


def retry_policy_provider(action, max_retries=2, reaction=None) -> RetryPolicy:
    return RetryPolicy(
        action, reaction=reaction, max_retries=max_retries, should_record=True
    )



from os import environ as envvars
from pathlib import Path
from setuptools import find_packages
from setuptools import setup


setup(
    name='recruitment',
    description='a lib for sourcing actionpacked entities capable of getting the job done',
    long_description=Path(__file__).absolute().parent.joinpath('README.md').read_text(),
    long_description_content_type='text/markdown',
    setup_requires=[
        'setuptools_scm==5.0.1'
    ],
    use_scm_version={'local_scheme': 'no-local-version'} if envvars.get('LOCAL_VERSION_SCHEME') else True,
    packages=find_packages(exclude=['tests']),
    author='Emmanuel I. Obi',
    maintainer='Emmanuel I. Obi',
    maintainer_email='withtwoemms@gmail.com',
    url='https://github.com/withtwoemms/recruitment',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires=[
        'actionpack==1.7.7',
        'boto3==1.21.10'
    ],
)

